#ifndef _STIM_DIAGRAM_GATE_DATA_3d_TEXTURE_DATA_H
#define _STIM_DIAGRAM_GATE_DATA_3d_TEXTURE_DATA_H

#include <string>

namespace stim_draw_internal {

std::string make_gate_3d_texture_data_uri();

}

#endif
